// section index is zero based (0,1,2,...,numofsections-1)
// ### HINT: hints

#define TPEFILE_FILESIZEMIN 98
//#define TPEFILE_FILESIZEMAX 50*1024*1024

typedef enum TPFERROR
{
	TPFE_NOERROR,
	TPFE_INVALIDPARAMETER,
	TPFE_INVALIDFILE,
	TPFE_CANTOPENFILE,
	TPFE_CANTALLOCLMEM,
	TPFE_CANTALLOCVMEM,
	TPFE_CANTREADFILE,
} TPFERROR;

#define TPFF_READONLY 1
#define TPFF_MEM 2
#define TPFF_MEMEX 4

///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
////
//// class: TPEFile
////
////
////
////
////
////

class TPEFile
{
	HANDLE hFile;
	BYTE *pMem;
	char *pszFileName;
	DWORD dwFileSize;
	DWORD dwFlags;
	TPFERROR LastErr;
	PIMAGE_DOS_HEADER pDosh;
	PIMAGE_NT_HEADERS pNth;
	PIMAGE_SECTION_HEADER pFirstSech; 

public:
	TPEFile()
	{
	
	};
	
	~TPEFile()
	{
	
	};

	BOOL LoadFile(char *pszPeFileName,BOOL bOpenForReading)
	{
		dwFlags=0;
		LastErr=TPFE_INVALIDPARAMETER;
		DWORD len=strlen(pszPeFileName),nbr;
		pszFileName=malloc(len+1);
		strcpy(pszFileName,pszPeFileName);
		dwFlags|=(bOpenForReading?TPFF_READONLY:0);
		LastErr=TPFE_CANTOPENFILE;
		hFile=CreateFile(pszPeFileName,\
		(bOpenForReading?GENERIC_READ:GENERIC_READ|GENERIC_WRITE),\
		FILE_SHARE_READ,0,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,0);
		if (hFile==INVALID_HANDLE_VALUE) goto ERR00;
		LastErr=TPFE_INVALIDFILE;
		dwFileSize=GetFileSize(hFile,0);
		if (dwFileSize<=TPEFILE_FILESIZEMIN) goto ERR01;
		LastErr=TPFE_CANTALLOCVMEM; 
		pMem=VirtualAlloc(0,dwFileSize,MEM_COMMIT,PAGE_READWRITE);
		if (pMem == NULL) goto ERR01;
		LastErr=TPFE_CANTREADFILE; 
		if (ReadFile(hFile,pMem,dwFileSize,&nbr,0)==0) goto ERR02;
		LastErr=TPFE_INVALIDFILE;
		pDosh=pMem;
		if (pDosh->e_magic!=IMAGE_DOS_SIGNATURE) goto ERR02;
		pNth=(PIMAGE_NT_HEADERS)((DWORD)pDosh+pDosh->e_lfanew);
		if (pNth->Signature!=IMAGE_NT_SIGNATURE) goto ERR02;
		pFirstSech=(PIMAGE_SECTION_HEADER)((DWORD)pNth+sizeof(IMAGE_NT_HEADERS));
		return 1;
		ERR02: VirtualFree(pMem,0,MEM_RELEASE);
		ERR01: CloseHandle(hFile);
		ERR00: free(pszFileName);
		return 0;
	}
	
	Reload()
	{
		char *pszTmp=malloc(strlen(pszFileName)+1);
		strcpy(pszTmp,pszFileName);
		TPEFile::Done();
		TPEFile::LoadFile(pszTmp,dwFlags&TPFF_READONLY);
		free(pszTmp);
	}
	
	BOOL Flush()
	{
		DWORD nbr;
		if (dwFlags&TPFF_READONLY) return 0;
		SetFilePointer(hFile,0,0,FILE_BEGIN);
		return (WriteFile(hFile,pMem,dwFileSize,&nbr,0)!=0);
	}

	BOOL SaveAs(char *pszPeFileName,BOOL bOverwrite)
	{
		HANDLE hDestFile; DWORD nbr;
		hDestFile=CreateFile(pszPeFileName,GENERIC_READ|GENERIC_WRITE,\
		0,0,bOverwrite?CREATE_ALWAYS:CREATE_NEW,FILE_ATTRIBUTE_NORMAL,0);
		if (hDestFile==INVALID_HANDLE_VALUE) return 0;
		nbr=WriteFile(hDestFile,pMem,dwFileSize,&nbr,0);
		CloseHandle(hDestFile);
		return nbr!=0;
	}
	
	Done()
	{
		VirtualFree(pMem,0,MEM_RELEASE);
		CloseHandle(hFile);
		free(pszFileName);
	}
	
	DWORD RvaToOffset(DWORD dwRva)
	{
		PIMAGE_SECTION_HEADER ish=pFirstSech;
		DWORD NumSec=pNth->FileHeader.NumberOfSections;
		for (;NumSec;ish++,NumSec--)
			if (
				(ish->VirtualAddress <= dwRva) && \
				(ish->VirtualAddress+ish->SizeOfRawData > dwRva) \
			)
				return dwRva-ish->VirtualAddress+ish->PointerToRawData;
		return 0;
	}
	
	DWORD OffsetToRva(DWORD dwOffs)
	{
		PIMAGE_SECTION_HEADER ish=pFirstSech;
		DWORD NumSec=pNth->FileHeader.NumberOfSections;
		for (;NumSec;ish++,NumSec--)
			if (
				(ish->PointerToRawData <= dwOffs) && \
				(ish->PointerToRawData+ish->Misc.VirtualSize > dwOffs) \
			)
				return dwOffs-ish->PointerToRawData+ish->VirtualAddress;
		return 0;
	}
	
	inline DWORD PtrFromRva(DWORD dwRva)
	{
		return (RvaToOffset(dwRva)+(DWORD)pMem);
	}

	inline DWORD PtrFromOffs(DWORD dwOffs)
	{
		return (dwOffs+(DWORD)pMem);
	}
	
	BYTE *DirEntryToData(int nDirectoryEntry,DWORD *pdwDataSize)
	{
		DWORD x=pNth->OptionalHeader.DataDirectory[nDirectoryEntry].Size;
		DWORD y=pNth->OptionalHeader.DataDirectory[nDirectoryEntry].VirtualAddress;
		if(y) y=PtrFromRva(y);
		if (pdwDataSize) *pdwDataSize=x; 
		return y;
	}
	
	DWORD FileAlign(DWORD dwValue)
	{
		DWORD x=dwValue;
		DWORD y=pNth->OptionalHeader.FileAlignment;
		return (x-1|y-1)+1;
	}
	
	DWORD MemAlign(DWORD dwValue)
	{
		DWORD x=dwValue;
		DWORD y=pNth->OptionalHeader.SectionAlignment;
		return (x-1|y-1)+1;
	}
	
	/* iSection = 0,1,2,..,NumSec-1 */
	BOOL IsRvaInSection(DWORD dwRva,int iSection)
	{
		DWORD NumSec=pNth->FileHeader.NumberOfSections;
		PIMAGE_SECTION_HEADER ish=pFirstSech;

		if (iSection<NumSec)
		{
			ish=(PIMAGE_SECTION_HEADER)((DWORD)ish+sizeof(IMAGE_SECTION_HEADER)*iSection);
			if (
				(ish->VirtualAddress <= dwRva) && \
				(ish->VirtualAddress+ish->SizeOfRawData > dwRva) \
				
			)
				return 1;
		}
		return 0;		
	}

	DWORD GetSectionType(int iSection)
	{
		DWORD Result=0,dirRva;
		for (int i=0; i<IMAGE_NUMBEROF_DIRECTORY_ENTRIES; i++ )
		{
			dirRva=pNth->OptionalHeader.DataDirectory[i].VirtualAddress;
			if (dirRva) Result|=(IsRvaInSection(dirRva,iSection)<<i);
		}
	}
	
	DWORD CalcChecksum()
	{
		WORD wTmpCSum=pNth->OptionalHeader.CheckSum;
		WORD wCarry=0,wCSum=pNth->OptionalHeader.CheckSum=0; 
		DWORD x,y=dwFileSize>>1;
		WORD *wp=(WORD *)((DWORD)pDosh+dwFileSize-2);
		while (y)
		{
			wCSum+=wCarry;
			wCarry=wCSum<wCarry;
			wCSum+=*wp;
			wCarry+=wCSum<*wp;
			wp--;
			y--;
		}
		x=(DWORD)wCSum+dwFileSize+(DWORD)wCarry;
		pNth->OptionalHeader.CheckSum=wTmpCSum;
		return x; 
	}
	
	BOOL AddSectionHdr(char *pSecName,DWORD dwVSize,DWORD dwRSize,DWORD dwSecFlags)
	{
		DWORD NumSec=pNth->FileHeader.NumberOfSections;
		// p = @ zadnja postojeca sekcija
		// psh = @ nova sekcija
		PIMAGE_SECTION_HEADER p=&pFirstSech[NumSec-1],psh=&pFirstSech[NumSec];
		// ima li mjesta?
		if (
			((DWORD)PtrFromOffs(pFirstSech->PointerToRawData) \
			-(DWORD)psh<sizeof(IMAGE_SECTION_HEADER))
			|| (NumSec==0)
		) return 0;
		memset(psh,0,sizeof(IMAGE_SECTION_HEADER));
		memcpy(&psh->Name[0],pSecName,8);
		psh->Misc.VirtualSize=MemAlign(dwVSize);
		psh->VirtualAddress=MemAlign(p->VirtualAddress+p->Misc.VirtualSize);
		psh->SizeOfRawData=FileAlign(dwRSize);
		psh->PointerToRawData=FileAlign(p->PointerToRawData+p->SizeOfRawData);
		psh->Characteristics=dwSecFlags;
		pNth->OptionalHeader.SizeOfImage=MemAlign(psh->VirtualAddress+psh->Misc.VirtualSize);
		pNth->FileHeader.NumberOfSections++;
		return 1;
	}

	BOOL DeleteSectionHdr(int iSection)
	{
		DWORD NumSec=pNth->FileHeader.NumberOfSections;
		// ### HINT: allow NumSec=0?
		if ( iSection>=NumSec || NumSec<=1 ) return 0;
		PIMAGE_SECTION_HEADER p=&pFirstSech[iSection];
		memset(p,0,sizeof(IMAGE_SECTION_HEADER));
		p--;// sekcija prije
		pNth->OptionalHeader.SizeOfImage=MemAlign(p->VirtualAddress+p->Misc.VirtualSize);
		pNth->FileHeader.NumberOfSections--;
		return 1;
	}
	
	BOOL DeleteLastSectionHdr()
	{
		return DeleteSectionHdr(pNth->FileHeader.NumberOfSections-1);
	}
	
	BOOL Rebase(DWORD dwNewBase)
	{
		PIMAGE_BASE_RELOCATION pr; WORD *pw;
		DWORD x,Chunks,Delta=dwNewBase-pNth->OptionalHeader.ImageBase;
		if (Delta==0) return 1;
		pw=DirEntryToData(IMAGE_DIRECTORY_ENTRY_BASERELOC,0);
		if (pw==NULL) return 0;
		pr=pw; pw=(WORD *)((DWORD)pw+sizeof(IMAGE_BASE_RELOCATION));
		while(pr->VirtualAddress)
		{
			Chunks=(pr->SizeOfBlock-sizeof(IMAGE_BASE_RELOCATION))>>1;
			while(Chunks)
			{
				x=(DWORD)*pw++;
				Chunks--;
				if (x&IMAGE_REL_BASED_HIGHLOW<<12)
				{
					x&=IMAGE_REL_BASED_HIGHLOW<<12-1;
					x+=pr->VirtualAddress;
					*((DWORD *)PtrFromRva(x))+=Delta;
				}
			}
		}
		return 1;
	}
	
	
};





///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
////
//// class: TPEImage
////
////
////
////
////
////

class TPEImage
{
	HINSTANCE hModule;
	BYTE *pMem;
	DWORD dwFileSize;
	DWORD dwFlags;
	DWORD dwPid;
	TPFERROR LastErr;
	PIMAGE_DOS_HEADER pDosh;
	PIMAGE_NT_HEADERS pNth;
	PIMAGE_SECTION_HEADER pFirstSech; 

public:
	TPEImage()
	{
	};
	
	~TPEImage()
	{
	};


	BOOL LoadFromAddress(BYTE *pAddress)// a DLL
	{
		BYTE *p=(DWORD)pAddress&0xFFFF0000;
		
		if (IsBadReadPtr(p,TPEFILE_FILESIZEMIN)!=0) return 0;
		pDosh=pMem=hModule=p;
		if (pDosh->e_magic!=IMAGE_DOS_SIGNATURE) goto ERR00;
		pNth=(PIMAGE_NT_HEADERS)((DWORD)pDosh+pDosh->e_lfanew);
		if (pNth->Signature!=IMAGE_NT_SIGNATURE) goto ERR00;
		pFirstSech=(PIMAGE_SECTION_HEADER)((DWORD)pNth+sizeof(IMAGE_NT_HEADERS));
		dwFlags=TPFF_MEM;
		return 1;
		
		ERR00:
		return 0;
	}
		
	BOOL LoadFromAddressEx(DWORD dwPid,BYTE *pAddress)
	{
		IMAGE_DOS_HEADER Dosh; IMAGE_NT_HEADERS Nth;
		HANDLE hProcess; BYTE *p=(DWORD)pAddress&0xFFFF0000;
		
		if (dwPid==GetCurrentProcessId()) return LoadFromAddress(pAddress);
		hProcess=OpenProcess(PROCESS_VM_READ,FALSE,dwPid);
		if (hProcess==NULL) goto ERR00;
		hModule=p;
		if (ReadProcessMemory(hProcess,p,&Dosh,sizeof(IMAGE_DOS_HEADER),0)==0) goto ERR00;
		if (Dosh.e_magic!=IMAGE_DOS_SIGNATURE) goto ERR00;
		if (ReadProcessMemory(hProcess,(BYTE *)((DWORD)p+Dosh.e_lfanew),\
			&Nth,sizeof(IMAGE_NT_HEADERS),0)==0) goto ERR00;
		if (Nth.Signature!=IMAGE_NT_SIGNATURE) goto ERR00;
		dwFileSize=Nth.OptionalHeader.SizeOfImage;
		pMem=VirtualAlloc(0,dwFileSize,MEM_COMMIT,PAGE_READWRITE);
		if (pMem == NULL) goto ERR00;
		if (ReadProcessMemory(hProcess,p,pMem,dwFileSize,0)==0) goto ERR01;
		pDosh=pMem;
		pNth=(PIMAGE_NT_HEADERS)((DWORD)pDosh+pDosh->e_lfanew);
		pFirstSech=(PIMAGE_SECTION_HEADER)((DWORD)pNth+sizeof(IMAGE_NT_HEADERS));
		dwFlags=TPFF_MEMEX;
		CloseHandle(hProcess);
		return 1;

		ERR01: VirtualFree(pMem,0,MEM_RELEASE);
		ERR00: CloseHandle(hProcess);
		return 0;
	}
	
	BOOL Reload()
	{
		TPEImage::Done();
		return TPEImage::LoadFromAddressEx(dwPid,hModule);
	}
	
	/*BOOL Flush()
	{
		HANDLE hProcess;
		
		if (dwFlags&TPFF_MEMEX)
		{
			hProcess=OpenProcess(PROCESS_VM_OPERATION,FALSE,dwPid);
			if (WriteProcessMemory(hProcess,hModule,pMem,dwFileSize,0))
			{
				return 1;
			}
			CloseHandle(hProcess);
		}
		else
		{
			if ((HMODULE)pMem!=GetModuleHandle(0))
			{
				return 1;
			}
			
		}
		return 0;
		
	}*/

	BOOL SaveAs(char *pszPeFileName,BOOL bOverwrite)
	{
		HANDLE hDestFile; DWORD nbr;
		hDestFile=CreateFile(pszPeFileName,GENERIC_READ|GENERIC_WRITE,\
		0,0,bOverwrite?CREATE_ALWAYS:CREATE_NEW,FILE_ATTRIBUTE_NORMAL,0);
		if (hDestFile==INVALID_HANDLE_VALUE) return 0;
		nbr=WriteFile(hDestFile,pMem,dwFileSize,&nbr,0);
		CloseHandle(hDestFile);
		return nbr!=0;
	}
	
	Done()
	{
		if ((dwFlags&TPFF_MEMEX)&&(pMem))
			VirtualFree(pMem,0,MEM_RELEASE);
	}


	DWORD RvaToOffset(DWORD dwRva)
	{
		PIMAGE_SECTION_HEADER ish=pFirstSech;
		DWORD NumSec=pNth->FileHeader.NumberOfSections;
		for (;NumSec;ish++,NumSec--)
			if (
				(ish->VirtualAddress <= dwRva) && \
				(ish->VirtualAddress+ish->SizeOfRawData > dwRva) \
			)
				return dwRva-ish->VirtualAddress+ish->PointerToRawData;
		return 0;
	}
	
	DWORD OffsetToRva(DWORD dwOffs)
	{
		PIMAGE_SECTION_HEADER ish=pFirstSech;
		DWORD NumSec=pNth->FileHeader.NumberOfSections;
		for (;NumSec;ish++,NumSec--)
			if (
				(ish->PointerToRawData <= dwOffs) && \
				(ish->PointerToRawData+ish->Misc.VirtualSize > dwOffs) \
			)
				return dwOffs-ish->PointerToRawData+ish->VirtualAddress;
		return 0;
	}	

	inline DWORD PtrFromRva(DWORD dwRva)
	{
		return (RvaToOffset(dwRva)+(DWORD)pMem);
	}

	inline DWORD PtrFromOffs(DWORD dwOffs)
	{
		return (dwOffs+(DWORD)pMem);
	}
	
	BYTE *DirEntryToData(int nDirectoryEntry,DWORD *pdwDataSize)
	{
		DWORD x=pNth->OptionalHeader.DataDirectory[nDirectoryEntry].Size;
		DWORD y=pNth->OptionalHeader.DataDirectory[nDirectoryEntry].VirtualAddress;
		if(y) y=PtrFromRva(y);
		if (pdwDataSize) *pdwDataSize=x; 
		return y;
	}
};


